require_relative "./escape.rb"
require_relative "./type.rb"
require_relative "./version.rb"

class Op < Struct.new(:name, :parse_nargs, :block, :ways)
  def dup
    Op.new(name.dup, parse_nargs, block, ways.map(&:dup))
  end
  def dup_mod(mod)
    return self if mod <= 0
    r = dup
    r.ways.each{|w| w.mod = mod }
    r
  end
end

class Way < Struct.new(:arg_ranks, :ret_rank, :arg_types, :ret_type, :mod, :vectorize, :impl)
  def mod_arg_ranks
    arg_ranks.map{|r| r < 0 ? ~r + mod : r }
  end
  def mod_ret_rank
    ret_rank < 0 ? ~ret_rank + mod : ret_rank
  end
  def dup
    Way.new(arg_ranks.dup, ret_rank, arg_types.dup, ret_type, mod, vectorize, impl)
  end
  def can_mod?
    (arg_ranks + [ret_rank]).any?{|r| r < 0 }
  end
end

def lookup_op(token)
  op = Ops[token.str] || Ops[token.name]
  return nil if !op
  if token.mod > 0
    if op.ways[0].can_mod?
      op = op.dup
      op.ways.each{|w|w.mod += token.mod }
    else
      op = Ops[token.str]
    end
  end
  op
end

def new_op(name, type_sig, block: false, vectorize: true, &impl)
  parts = type_spec_parts(type_sig)
  *arg_ranks, ret_rank = parts.map{|arg| spec_rank(arg) }
  *arg_types, ret_type = parts.map{|arg| spec_type(arg) }

  way = Way.new(arg_ranks, ret_rank, arg_types, ret_type, 0, vectorize, impl)
  Op.new(name, arg_ranks.size, block, [way])
end

Ops = {}
OpExamples = {}

if !defined? mk_op
  def mk_op(names, type_sig, example="", block: false, vectorize: true, &impl)
    OpExamples[names] = example if example != ""
    names = names.split

    op = new_op(names[-1], type_sig, block: block, vectorize: vectorize, &impl)

    names.each{|name|
      if Ops[name]
        Ops[name].ways << op.ways[0]
      else
        Ops[name] = op.dup
      end
    }

    if op.ways[0].can_mod?
      mod_op = op.dup
      mod_op.ways[0].mod = 1

      names.each{|name|
        if name =~ /^[a-z]/
          name = name.sub(/^./){|first|first.upcase}
          if Ops[name]
            Ops[name].ways << mod_op.ways[0]
          else
            Ops[name] = mod_op.dup
          end
        end
      }
    end
  end

  def category(name); end
  def quickref_op(names, desc, example)
    OpExamples[names] = example if example != ""
  end
end

NoPromote = ->i{i}

category("mathy ops")
mk_op("+ add", "int int -> int", '1 2+ -> 3', &add=->a,b{ a.value+b.value })
mk_op("+ add", "int char -> char", "1 'a+ -> 'b", &add)
mk_op("+ add", "char int -> char", "'a 1+ -> 'b", &add)
mk_op("_ sum", "[int] -> int", '1,2,3_ -> 6', &sum=->a{ to_strict_list(a).sum })
mk_op("_ sum", "[char] -> int", '"ab"_ -> 195', &sum)
mk_op(". prod", "[int] -> int", '1,2,3. -> 6'){|a| to_strict_list(a).inject(1,&:*) }
mk_op("- sub", "int int -> int", '5 3- -> 2', &subt=->a,b{ a.value-b.value })
mk_op("- sub", "char int -> char", "'b 1- -> 'a", &subt)
mk_op("- sub", "char char -> int", "'b 'a- -> 1", &subt)
mk_op("~ negate", "int -> int", '1~ -> -1'){|a| -a.value }
mk_op("* mult", "int int -> int", '2 3* -> 6'){|a, b| a.value*b.value }
mk_op("/ div", "int int -> int", '7 2/ -> 3'){|a, b| b.value == 0 ? 0 : a.value/b.value }
mk_op("% mod", "int int -> int", '7 2% -> 1'){|a, b| b.value == 0 ? 0 : a.value%b.value }
mk_op("^ pow", "int int -> int", '2 3^ -> 8'){|a, b| a.value == 0 && b.value < 0 ? 0 : (a.value**b.value).to_i }
mk_op("` str", "int -> [char]", '5` -> "5"'){|a| str(a) }
mk_op("& chr", "int -> char", "97& -> 'a"){|a| a.value }
mk_op("# #, toBase", "int int -> [int]", '6 2# -> [1,1,0]'){|a,b| to_base(a.value,b.value) }
mk_op("# fromBase", "[int] int -> int", '1,1,0 2# -> 6'){|a,b| from_base(a.value,b.value) }
mk_op("| abs", "int -> int", '5~| -> 5'){|a| a.value.abs }
mk_op("( pred", "int -> int", '4( -> 3', &pred=->a { a.value-1 })
mk_op("( pred", "char -> char", "'b( -> 'a", &pred)
mk_op(") succ", "int -> int", '4) -> 5', &succ=->a { a.value+1 })
mk_op(") succ", "char -> char", "'a) -> 'b", &succ)
mk_op("} countTo", "int -> [int]", "3} -> [1,2,3]"){|a| range(1,a.value+1) }
mk_op("{ wholes", "[int]", "{ => [0,1,2,..."){ range_from(0) }

category("char ops")
mk_op(". read", "[char] -> int", '"23". -> 23'){|a| read_num(a)[0] }
mk_op("~ readAll", "[char] -> [int]", '"1,2 3"~ -> [1,2,3]'){|a| split_non_digits(a) }
mk_op("| ord", "char -> int", "'a| -> 97"){|a| a.value }
mk_op("` strip", "[char] -> [char]", '" ab\n"` -> "ab"'){|a| strip(a,$at,$rank) }
mk_op("* retell", "[char] int -> [char]", '"ab"3* -> "ababab"'){|a, b|
  concat(take(b.value,repeat(a).const).const) }
mk_op("* join", "[[char]] [char] -> [char]", '1,2,3 " "* -> "1 2 3"'){|a, b| join(a,b)}
mk_op("/ split", "[char] [char] -> [[char]]", '"a  b c " " "/ -> ["a","b","c"]'){|a, b| s=split(a,b).const; filter(s,s,nil,1) }
mk_op("% cut", "[char] [char] -> [[char]]",'"a  b c " " "% -> ["a","","b","c",""]'){|a, b| split(a,b) }

category("generic ops")
mk_op("a append", "[a] [a] -> [a]", '"ab" "cd"a -> "abcd"'){|a, b| append(a,b) }
mk_op("b backwards reverse", "[a] -> [a]", '"abc"b -> "cba"'){|a| reverse(a) }
mk_op("c cons", "[a] a -> [a]", '1,2 0c -> [0,1,2]'){|a, b| [b, a] }
mk_op("d drop", "[a] int -> [a]", '"abcde" 2d -> "cde"'){|a, b| drop(b.value, a) }
mk_op("g groupWhile", "[a] [b] -> [[a]]", '1,2,3,4,5:3%g -> [[1,2],[4,5]]'){|a,b| chunk_while(a,b,$rank,$bt) }
mk_op("h head", "[a] -> a", '"abc"h -> \'a'){|a| z=zero($rank,$at); a.empty ? z : a.value[0].value }
mk_op("j just", "a -> [a]", "1j -> [1]"){|a|[a, [].const]}
mk_op("k keep take", "[a] int -> [a]", '"abcde" 2k -> "ab"'){|a, b| take(b.value, a) }
mk_op("l len", "[a] -> int", '"abc"l -> 3'){|a| len(a.value) }
mk_op("n not", "a -> int", '0,1,2n -> [1,0,0]'){|a| truthy($at, $rank, a.value) ? 0 : 1 }
mk_op("O Or o, or,", "int [char] -> [char]"){|a,b| truthy($at, 0, a.value) ? str(a) : b.value }
mk_op("o or", "a a -> a", '" bc" \'- o -> "-bc"'){|a,b| truthy($at, $rank, a.value) ? a.value : b.value }
mk_op("p priorCount tally", "[a] -> [int]", '"aardvark"p -> [0,1,0,0,0,2,1,0]'){|a| tally(a) }
mk_op("q equal", "a a -> int", "1,2,3 1q -> [1,0,0]"){|a,b| spaceship(a, b, $rank) == 0 ? 1 : 0 }
mk_op("r repeat", "a -> [a]", "1r => 1,1,1,1,1,1..."){|a| repeat(a) }
mk_op("s sortBy", "[a] [b] -> [a]", '"abc","d":len S -> ["d","abc"]'){|a,b| sort_by(a,b,$rank) }
mk_op("t tail", "[a] -> [a]", '"abc"t -> "bc"'){|a| a.empty ? [] : a.value[1].value }
mk_op("u unite concat", "[[a]] -> [a]", '"ab","cd","e"u -> "abcde"'){|a| concat(a) }
mk_op("v vet filter", "[a] [b] -> [a]", '"abcdef":|2%v -> "ace"'){|a,b| filter(a,b,$bt,$rank) }
mk_op("w while takeWhile", "[a] [b] -> [a]", '"abc def":w -> "abc"'){|a,b| take_while(a,b,$bt,$rank) }
mk_op("x exclude setDiff", "[a] [a] -> [a]", '"abra" "ra"x -> "ba"'){|a,b| set_diff(a,b) }
mk_op("y and", "a b -> b", '" ab" "123"y -> " 23"'){|a,b| z=zero($rank,$bt); truthy($at, $rank, a.value) ? b.value : z }
mk_op("z zenith last", "[a] -> a", '"abc"z -> \'c'){|a| z=zero($rank,$at); last(a, z.const) }
mk_op("< lessThan", "a a -> int", '1,2,3 2< -> [1,0,0]'){|a,b| spaceship(a,b,$rank) == -1 ? 1 : 0 }
mk_op("? reshape", "[a] int -> [[a]]", '"abcdef" 2? -> ["ab","cd","ef"]'){|a,b| reshape(a,b.value) } # todo move in quickref
mk_op("\\ transpose", "[[a]] -> [[a]]", '"abc","def"\\ -> ["ad","be","cf"]'){|a| transpose(a) }


category("folding ops") # (return type is type yielded)
mk_op("i iterate", "a -> [a]", '0 i 1,2,3+ -> [0,1,3,6]', block: true)
mk_op("e expand iterate0", "[a]", 'e 1,2,3+ -> [0,1,3,6]', block: true)
mk_op("f foldr", "a -> [a]", '0 f 1,2,3+ -> 6', block: true)
mk_op("m meld foldr0", "[a]", 'm 1,2,3+ -> 6', block: true)

category("stack ops")
mk_op(": dup", "a -> a", "5: -> 5 5")
mk_op("; mdup", "a -> a", "5;1- -> 4 5", block: true)

category("special symbols")
quickref_op("$", "arg", "5;$+ -> 10 5")
quickref_op("!", "outer consume", "2 5;!+ -> 7 5")
quickref_op(">", "end block", "3;))> -> 5 3")
quickref_op(",", "unvec", '"ab"j, -> ["ab"]')
quickref_op("@", "register", "@( 5@) -> 4 6")
quickref_op("[ ]", "push pop", "5[( ]) -> 4 6")
quickref_op("=", "assignment", "5=a( a) -> 4 6")

category("ops without short name")
mk_op("stdin", "[char]", "<stdin>"){ ReadStdin.value }
mk_op("!, undefined", "a", '!, => error: undefined'){ raise IogiiError.new "undefined" }
mk_op("err error", "[char] -> b", '"msg" err => error: msg'){|a| raise IogiiError.new to_eager_str(a.value) }

category("debug ops (short version pops and prints)") # (unvectorized)
mk_op("ppi inspect", "a -> [char]", '5 inspect -> "5"', vectorize: false){|a| inspectv($at, $rank, a) }
mk_op("ppt type", "a -> [char]", '5 type -> "int"', vectorize: false){|a| str_to_lazy_list(type_and_rank_to_str($at, $rank)) }
mk_op("version", "[char]", "version -> \"#{Version}\""){ str_to_lazy_list(Version) }

category("low rank overloads")
mk_op(". ., pow10", "int -> int", "2. -> 100"){|a| (10**a.value).to_i }
mk_op("B countToPow10", "int -> [int]", "2B => [1,2,...,100]"){|a| range(1,(10**a.value).to_i+1) }
mk_op("? replicate", "int a -> [a]", '4 3? -> [3,3,3,3]'){|a,b| take(a.value,repeat(b).const) }
mk_op("\\ digits", "int -> [int]", '123 \\ -> [1,2,3]'){|a| to_base(a.value,10) }
mk_op("U undigits", "[int] -> int", '1,2,3 U -> 123'){|a| from_base(a.value,10) }

category("deprecated ops") # or internally useful ops
mk_op("lines", "[char] -> [[char]]"){|a| lines(a) }
mk_op("greaterThan", "a a -> int"){|a,b| spaceship(a,b,$rank) == 1 ? 1 : 0 }
mk_op("uniq", "[a] -> [a]"){|a| filter(a,Promise.new{isuniq(a)},IntType,0)}
mk_op("pair", "a a -> [a]"){|a, b| [a, [b, Null].const] }
mk_op("pad", "[a] a -> [a]"){|a, b| padv(a, b) }
mk_op("init", "[a] -> [a]"){|a| init(a) }
mk_op("tails", "[a] -> [[a]]"){|a| tails(a) }
mk_op("min", "a a -> a"){|a,b| spaceship(a,b,$rank) <= 0 ? a.value : b.value }
mk_op("max", "a a -> a"){|a,b| spaceship(a,b,$rank) >= 0 ? a.value : b.value }
mk_op("sort", "[a] -> [a]"){|a|sort(a,$rank) }
mk_op("ifelse", "a b b -> b"){|a,b,c| truthy($at, $rank, a.value) ? b.value : c.value }
mk_op("nil", "[a]"){ [] }
mk_op("superpad", "[a] -> [a]"){ raise }
mk_op("zero", "[] -> int"){|a| 0 } # default to int in fold0s
mk_op("zerom", "[a] -> a"){|a| zero($rank,$at) } # this is used for rank1+ fold0s
mk_op("zero", "[a] -> a"){|a| zero($rank,$at) }

category("intuitive special promotes (don't need to show them in quickref")
mk_op("* join", "[char] [char] -> [char]"){|a, b| join(map(a){|i| [i,Null] }.const,b)}

mk_op("g groupWhile", "[a] b -> [[a]]"){|a,b| chunk_while(a,b,$rank-1,$bt) }
mk_op("s sortBy", "[a] b -> [a]"){|a,b| sort_by(a,b,$rank-1) }
mk_op("s sortBy", "a b -> a", &NoPromote)
mk_op("v vet filter", "[a] b -> [a]"){|a,b| filter(a,b,$bt,$rank-1) }
mk_op("w takeWhile", "[a] b -> [a]"){|a,b| take_while(a,b,$bt,$rank-1) }

mk_op("* join", "char [char] -> [char]", &NoPromote)
mk_op("_ sum", "int -> int", &NoPromote)
mk_op("prod", "int -> int", &NoPromote)
mk_op("reshape", "a int -> [[a]]", &NoPromote)
mk_op("transpose", "[a] -> [[a]]", &NoPromote)
mk_op("unite concat", "[a] -> [a]", &NoPromote)
mk_op("u", "[int] -> [int]", &NoPromote) # don't use generic for def since would override U (todo still true?)
mk_op("u", "[char] -> [char]", &NoPromote)
mk_op("U undigits", "int -> int", &NoPromote)
mk_op("h head", "a -> a", &NoPromote)
mk_op("z zenith last", "a -> a", &NoPromote)
# actually useful for nil replacement: mk_op("t tail", "a -> [a]", &NoPromote)
mk_op("backwards reverse", "a -> [a]", &NoPromote) # todo what about no promote for b?
mk_op("l len", "a -> int", &NoPromote)
mk_op("p priorCount tally", "a -> [int]", &NoPromote)
mk_op("superpad", "a -> [a]", &NoPromote)

#capital versions todo (maybe more)
#mk_op("n not", "a -> int"){|a| truthy($at, $rank, a.value) ? 0 : 1 }
